/* Distributed Checksum Clearinghouse heap debugging
 *
 * Copyright (c) 2004 by Rhyolite Software
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND RHYOLITE SOFTWARE DISCLAIMS ALL
 * WARRANTIES WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL RHYOLITE SOFTWARE
 * BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
 * SOFTWARE.
 *
 * Rhyolite Software DCC 1.2.36-1.8 $Revision$
 */

#include "dcc_defs.h"
#include "dcc_heap_debug.h"
#include <stdio.h>

#ifdef UNIX
#include <unistd.h>
#if defined(FreeBSD) || defined(linux)
#define NO_MALLOC_OPTIONS
#endif
char dcc_malloc_options[] = "A";
#else /* !UNIX */
#include "malloc.h"
#endif /* !UNIX */


#define SENTINEL_VALUE 0xdeadbeaf
typedef u_long SENTINEL;
SENTINEL tail = SENTINEL_VALUE;

typedef struct mdbg{
    SENTINEL	head;
    struct mdbg *fwd, *bak;
    u_char	*tail;
} MDBG;

u_int dcc_num_mdbg;
MDBG *dcc_mdbg_chain;

/* leave a few names outside to make some `ar`s happy */
#ifdef DEBUG_HEAP


/* this needs to do as little as possible to avoid calling malloc() */
void NRATTRIB
dcc_heap_abort(const char *m)
{
#ifdef UNIX
	write(STDERR_FILENO, m, strlen(m));
	abort();
#else
	bad_message_box("heap_debug", 1, m);
#endif
}



char *
dcc_strdup(const char *s)
{
	u_int len;
	char *p;

	len = strlen(s)+1;
	p = dcc_malloc(len);
	if (len > 1)
		memcpy(p, s, len);
	else
		p[len-1] = '\0';
	return p;
}



void
dcc_malloc_check(void)
{
	MDBG *mp;
	u_int i;

	i = dcc_num_mdbg;
	if (!i)
		return;

	mp = dcc_mdbg_chain;
	for (;;) {
		if (mp->head != SENTINEL_VALUE
		    || memcmp(mp->tail, &tail, sizeof(tail)))
			dcc_heap_abort("trashed heap sentinel");
		if (mp->bak->fwd != mp
		    || mp->fwd->bak != mp)
			dcc_heap_abort("malloc chain trashed");
		if (--i == 0) {
			if (mp->fwd != dcc_mdbg_chain)
				dcc_heap_abort("wrong malloc chain too long");
#ifndef UNIX
			i = _heapchk();
			if (i != _HEAPOK && i != _HEAPEMPTY)
				dcc_heap_abort("heapchk() failed");
#endif
			return;
		} else if (mp->fwd == dcc_mdbg_chain) {
			dcc_heap_abort("wrong malloc chain too short");
		}
		mp = mp->fwd;
	}
}



void *
dcc_malloc(size_t len)
{
	MDBG *mp;

#if defined(UNIX) && !defined(NO_MALLOC_OPTIONS)
	if (!malloc_options) {
		malloc_options = dcc_malloc_options;
	} else if (malloc_options != dcc_malloc_options) {
		dcc_heap_abort("malloc_options already set");
	}
#endif

	dcc_malloc_check();

	if (!len)
		dcc_heap_abort("malloc(0)");

	len += sizeof(MDBG) + sizeof(tail);

	mp = malloc(len);
	if (!mp)
		return mp;

	if (!dcc_num_mdbg) {
		mp->fwd = mp->bak = mp;
	} else {
		mp->bak = dcc_mdbg_chain;
		mp->fwd = dcc_mdbg_chain->fwd;
		mp->bak->fwd = mp;
		mp->fwd->bak = mp;
	}
	dcc_mdbg_chain = mp;
	mp->head = SENTINEL_VALUE;
	mp->tail = (u_char *)mp+len-sizeof(tail);
	memcpy(mp->tail, &tail, sizeof(tail));
	dcc_num_mdbg++;

	return (mp+1);
}



void *
dcc_calloc(size_t n, size_t s)
{
	void *p;

	s *= n;
	if (s == 0)
		dcc_heap_abort("zero calloc() size");
	p = dcc_malloc(s);
	if (!p)
		return p;
	memset(p, 0, s);
	return p;
}



void
dcc_free(void *p)
{
	int i;
	MDBG *mp;

	dcc_malloc_check();
	i = dcc_num_mdbg;
	mp = dcc_mdbg_chain;
	for (;;) {
		if (!i)
			dcc_heap_abort("freeing non-free");
		if (mp+1 == p)
			break;
		mp = mp->fwd;
		i--;
	}

	if (dcc_mdbg_chain == mp)
		dcc_mdbg_chain = mp->fwd;
	mp->bak->fwd = mp->fwd;
	mp->fwd->bak = mp->bak;
	dcc_num_mdbg--;
	memset(mp, 0xf1, mp->tail+sizeof(tail)-(u_char *)mp);
	dcc_malloc_check();

	free(mp);
}
#endif /* DEBUG_HEAP */
